@@ -4,6 +4,7 @@ from typing import Iterable
 
 import matplotlib.pyplot as plt
 import numpy as np
+from copy import copy
 
 from ._filters import filterplot
 from ._general import imgplot
@@ -14,20 +15,20 @@ __all__ = ["ParamGrid", "ImageGrid", "rgbplot", "FilterGrid"]
 
 class ImageGrid:
     """
-    Figure level : plot a collection of 2-D images or 3-D image data
-    along a grid. This class also supports slicing of the 3-D image
+    Figure level : plot a collection of 2-D or 3-D images or 3-D or 4-D image data
+    along a grid. This class also supports slicing of the 3-D and 4-D image data
     along different axis with variable step sizes and start/end indexes.
 
     Parameters
     ----------
     data :
-        3-D Image data (array-like) or list of 2-D image data. Supported array shapes are all
+        3-D or 4-D Image data (array-like), or list of 2-D or 3-D image data. Supported array shapes are all
         `matplotlib.pyplot.imshow` array shapes
     slices : int or list, optional
-        If `data` is 3-D, `slices` will index the specific slice from the last axis and only plot
+        If `data` is 3-D or 4-D, `slices` will index the specific slice from the last axis and only plot
         the resulting images. If None, it will plot all the slices from the last axis, by default None
     axis : int, optional
-        Axis along which the data will be sliced, by default -1
+        Axis along which the data will be sliced, by default -1 for 3-D arrays and 0 for 4-D arrays
     step : int, optional
         Step along the given axis, by default 1
     start : int, optional
@@ -42,8 +43,8 @@ class ImageGrid:
         Number of columns to display. Defaults to None.
     height : int or float, optional
         Size of the individual images. Defaults to 3.
-    aspect : int or float, optional
-        Aspect ratio of individual images. Defaults to 1.
+    aspect : int, float or 'auto', optional
+        Aspect ratio of individual images, when set to 'auto', it calculates the aspect ratio of the images passed. Defaults to 'auto'.
     cmap : str or `matplotlib.colors.Colormap` or list, optional
         Image colormap. If input data is a list of images,
         `cmap` can be a list of colormaps. Defaults to None.
@@ -56,6 +57,9 @@ class ImageGrid:
         on the percentiles specified instead of the extremes, by default (2, 98) -
         2nd and 98th percentiles for min and max values. Can be a list of tuples, if
         input data is a list of images
+    diverging : bool or list, optional
+        If True, vmax and vmin are adjusted so they have the same absolute value, making the diverging
+        color maps show 0 at the middle.
     alpha : float or array-like, optional
         `matplotlib.pyplot.imshow` alpha blending value from 0 (transparent) to 1 (opaque),
         by default None
@@ -67,6 +71,9 @@ class ImageGrid:
         Maximum data value that colormap covers, by default None
     interpolation : str, optional
         `matplotlib.pyplot.imshow` interpolation method used, by default None
+    norm : `matplotlib.colors.Normalize` or list of `matplotlib.colors.Normalize`, optional
+        `matplotlib` Normalize instance used to scale scalar data before
+        mapping to colors using cmap.
     dx : float or list, optional
         Size per pixel of the image data. If scalebar
         is required, `dx` and `units` must be sepcified.
@@ -109,6 +116,9 @@ class ImageGrid:
     despine : bool, optional
         Remove axes spines from image axes as well as colorbar axes.
         Defaults to None.
+    extent : floats (left, right, bottom, top), optional
+        The bounding box in data coordinates that the image will fill.
+        The image is stretched individually along x and y to fill the box.
 
     Returns
     -------
@@ -119,15 +129,17 @@ class ImageGrid:
     ValueError
         If `data` is None
     ValueError
-        If `data` has more than 3 dimensions
+        If `data` has 1 dimension
     ValueError
-        If `data` contains a 3D image within a list of images
+        If `data` has more than 4 dimensions
+    ValueError
+        If `data` contains a 4D image within a list of images
     ValueError
         If `axis` is not 0, 1, 2 or -1
     TypeError
         If `map_func` is not a callable object or a list/tuple of callable objects
     ValueError
-        If `map_func` is a list/tuple of callable objects when `data` is 3D
+        If `map_func` is a list/tuple of callable objects when `data` is 3D or 4D
 
     Examples
     --------
@@ -214,6 +226,22 @@ class ImageGrid:
 
         >>> g = isns.ImageGrid(cells, vmin=0, vmax=1, height=1, col_wrap=5)
 
+    Plot a list of 3-D images
+
+    .. plot::
+        :context: close-figs
+
+        >>> from skimage.data import astronaut, chelsea
+        >>> g = isns.ImageGrid([astronaut(), chelsea()], origin="upper")
+
+    Plot 4-D image data cube
+
+    .. plot::
+        :context: close-figs
+
+        >>> cifar = isns.load_image("cifar10")
+        >>> g = isns.ImageGrid(cifar, height=1, col_wrap=6, origin="lower")
+
     Map a function to the image data
 
     .. plot::
@@ -262,7 +290,8 @@ class ImageGrid:
         ...             [pol, pl, retina],
         ...             map_func=[gaussian, median, hessian],
         ...             dx=[15, 100, None],
-        ...             units="nm")
+        ...             units="nm",
+        ...             aspect=1)
 
     Change colorbar orientation
 
@@ -285,7 +314,7 @@ class ImageGrid:
         data,
         *,
         slices=None,
-        axis=-1,
+        axis=None,
         step=1,
         start=None,
         stop=None,
@@ -293,15 +322,17 @@ class ImageGrid:
         map_func_kw=None,
         col_wrap=None,
         height=3,
-        aspect=1,
+        aspect="auto",
         cmap=None,
         robust=False,
         perc=(2, 98),
+        diverging=False,
         alpha=None,
         origin=None,
         vmin=None,
         vmax=None,
         interpolation=None,
+        norm=None,
         dx=None,
         units=None,
         dimension=None,
@@ -312,6 +343,7 @@ class ImageGrid:
         cbar_ticks=None,
         showticks=False,
         despine=None,
+        extent=None,
     ):
         if data is None:
             raise ValueError("image data can not be None")
@@ -322,7 +354,7 @@ class ImageGrid:
             # check the number of images to be plotted
             _nimages = len(data)
 
-            # --- List/Tuple of 2D images with a List/Tuple of map_func ---
+            # --- List/Tuple of 2D or 3D images with a List/Tuple of map_func ---
             # change the number of images on the grid accordingly
             map_func_type = self._check_map_func(map_func, map_func_kw)
             if map_func_type == "list/tuple":
@@ -334,13 +366,52 @@ class ImageGrid:
                         len(map_func) if len(map_func) >= len(data) else len(data)
                     )
 
-        elif data.ndim > 3:
-            raise ValueError("image data can not have more than 3 dimensions")
+            if aspect == "auto":
+                aspect_ratios = [img.shape[1] / img.shape[0] for img in data]
+                aspect = min(aspect_ratios)
+
+        elif not isinstance(data, np.ndarray):
+            raise ValueError("image data must be a list of images or a 3d or 4d array.")
+
+        elif data.ndim == 2:
+            warnings.warn(
+                "The data inputed is a 2d array which contains a single image. "
+                "It is recomended that you use `imgplot` instead of `ImageGrid`.",
+                RuntimeWarning,
+            )
+            _nimages = 1
+
+            # ---- 2D image with a list/tuple of map_func or individual map_func ------
+            # check if map_func is a list/tuple of callables
+            # and assign the new number of images
+            map_func_type = self._check_map_func(map_func, map_func_kw)
+            if map_func_type == "list/tuple":
+                _nimages = len(map_func)
+                # no of columns should now be len of map_func list
+                col_wrap = len(map_func) if col_wrap is None else col_wrap
+
+            if aspect == "auto":
+                aspect = data.shape[1] / data.shape[0]
+
+        elif data.ndim in [3, 4]:
+            if data.ndim == 4 and data.shape[-1] not in [1, 3, 4]:
+                raise ValueError(
+                    "The number of channels in the images must be 1, 3 or 4"
+                )
+
+            if axis is None:
+                if data.ndim == 3:
+                    axis = -1
+                else:
+                    axis = 0
+
+            if data.ndim == 3 and axis not in [0, 1, 2, -1]:
+                raise ValueError("Incorrect 'axis'; must be either 0, 1, 2, or -1")
+
+            if data.ndim == 4 and axis not in [0, 1, 2, 3, -1]:
+                raise ValueError("Incorrect 'axis'; must be either 0, 1, 2, 3, or -1")
 
-        elif data.ndim == 3:
             if slices is None:
-                if axis not in [0, 1, 2, -1]:
-                    raise ValueError("Incorrect 'axis'; must be either 0, 1, 2, or -1")
                 # slice the image array along specified axis;
                 # if start, stop and step are not provided, default is step=1
                 data = data[
@@ -355,34 +426,30 @@ class ImageGrid:
 
             _nimages = len(slices)
 
-            # ---- 3D image with an individual map_func ----
+            if aspect == "auto":
+                # Select axis where width and height are
+                height_idx, width_idx = [i for i in range(data.ndim) if i != (axis % data.ndim)][:2]
+
+                aspect = data.shape[width_idx] / data.shape[height_idx]
+
+            # ---- 3D or 4D image with an individual map_func ----
             map_func_type = self._check_map_func(map_func, map_func_kw)
             # raise a ValueError if a list of map_func is provided for 3d image
             # TODO - support multiple map_func if "chaining"?
             if map_func_type == "list/tuple":
                 raise ValueError(
-                    "Can not map multiple functions to a 3D image. Please provide a single `map_func`"
+                    "Can not map multiple functions to 3D or 4D image. Please provide a single `map_func`"
                 )
 
         else:
-            # if data dim is not >2,
-            _nimages = 1
-
-            # ---- 2D image with a list/tuple of map_func or individual map_func ------
-            # check if map_func is a list/tuple of callables
-            # and assign the new number of images
-            map_func_type = self._check_map_func(map_func, map_func_kw)
-            if map_func_type == "list/tuple":
-                _nimages = len(map_func)
-                # no of columns should now be len of map_func list
-                col_wrap = len(map_func) if col_wrap is None else col_wrap
+            raise ValueError("Image data can not have more than 4 dimensions")
 
         # if no column wrap specified
         # set it to default 3
         if col_wrap is None:
             col_wrap = 3
 
-            # don't create extra columns when there aren't enough images
+        # don't create extra columns when there aren't enough images
         if col_wrap > _nimages:
             col_wrap = _nimages
 
@@ -412,11 +479,13 @@ class ImageGrid:
         self.cmap = cmap
         self.robust = robust
         self.perc = perc
+        self.diverging = diverging
         self.alpha = alpha
         self.origin = origin
         self.vmin = vmin
         self.vmax = vmax
         self.interpolation = interpolation
+        self.norm = norm
         self.dx = dx
         self.units = units
         self.dimension = dimension
@@ -427,6 +496,7 @@ class ImageGrid:
         self.cbar_ticks = cbar_ticks
         self.showticks = showticks
         self.despine = despine
+        self.extent = extent
 
         self._nrow = nrow
         self._ncol = ncol
@@ -477,8 +547,10 @@ class ImageGrid:
         _cmap = self.cmap
         _robust = self.robust
         _perc = self.perc
+        _diverging = self.diverging
         _vmin = self.vmin
         _vmax = self.vmax
+        _norm = self.norm
         _dx = self.dx
         _units = self.units
         _dimension = self.dimension
@@ -493,23 +565,19 @@ class ImageGrid:
                 _d = self.data[i]
 
                 # check if the image has more than 2 dimensions
-                if _d.ndim > 2:
+                if _d.ndim > 3:
                     raise ValueError(
-                        "can not plot multiple 3D images. Supply an individual 3D image"
+                        f"Image {i} in the list has more than 3 dimensions"
                     )
 
-            elif self.data.ndim == 3:
-                if self.axis == 0:
-                    _d = self.data[self.slices[i], :, :]
-                elif self.axis == 1:
-                    _d = self.data[:, self.slices[i], :]
-                elif self.axis == 2 or self.axis == -1:
-                    _d = self.data[:, :, self.slices[i]]
+                if _d.ndim == 3 and _d.shape[-1] not in [1, 3, 4]:
+                    raise ValueError(f"Image {i} in the list has more than 4 channels")
+
+            elif self.data.ndim == 2:
+                _d = self.data
 
             else:
-                # if a single 2D image is supplied
-                # TODO issue a warning and direct the user to imgplot()
-                _d = self.data
+                _d = self.data.take(indices=self.slices[i], axis=self.axis)
 
             if isinstance(self.cmap, (list, tuple)):
                 self._check_len_wrt_n_images(self.cmap)
@@ -519,6 +587,10 @@ class ImageGrid:
                 self._check_len_wrt_n_images(self.robust)
                 _robust = self.robust[i]
 
+            if isinstance(self.diverging, (list, tuple)):
+                self._check_len_wrt_n_images(self.diverging)
+                _diverging = self.diverging[i]
+
             if isinstance(self.vmin, (list, tuple)):
                 self._check_len_wrt_n_images(self.vmin)
                 _vmin = self.vmin[i]
@@ -531,6 +603,10 @@ class ImageGrid:
                 self._check_len_wrt_n_images(self.perc)
                 _perc = self.perc[i]
 
+            if isinstance(self.norm, (list)):
+                self._check_len_wrt_n_images(self.norm)
+                _norm = self.norm[i]
+
             if isinstance(self.dx, (list, tuple)):
                 self._check_len_wrt_n_images(self.dx)
                 _dx = self.dx[i]
@@ -561,11 +637,13 @@ class ImageGrid:
                 cmap=_cmap,
                 robust=_robust,
                 perc=_perc,
+                diverging=_diverging,
                 vmin=_vmin,
                 vmax=_vmax,
                 alpha=self.alpha,
                 origin=self.origin,
                 interpolation=self.interpolation,
+                norm=_norm,
                 dx=_dx,
                 units=_units,
                 dimension=_dimension,
@@ -576,6 +654,7 @@ class ImageGrid:
                 cbar_ticks=self.cbar_ticks,
                 showticks=self.showticks,
                 despine=self.despine,
+                extent=self.extent,
                 describe=False,
             )
 
@@ -626,7 +705,8 @@ class ImageGrid:
 
     def _map_func_to_data(self, map_func, map_func_kw):
         """Transform image data using the map_func callable object."""
-        # if data is a list or tuple of 2D images
+        self.data = copy(self.data)
+        # if data is a list or tuple of 2D or 3D images
         if isinstance(self.data, (list, tuple)):
             if self._check_map_func(map_func, map_func_kw) == "list/tuple":
                 self._adjust_param_list_len(map_func)
@@ -651,15 +731,15 @@ class ImageGrid:
                     else:
                         self.data[i] = map_func(self.data[i])
 
-        # if data is 3D or 2D and map_func is single callable
         else:
+            # if data is 4D, 3D or 2D and map_func is single callable
             if self._check_map_func(map_func, map_func_kw) == "callable":
-                # only pass on kwargs if not None
                 if map_func_kw is not None:
                     self.data = map_func(self.data, **map_func_kw)
                 else:
                     self.data = map_func(self.data)
-            # list of callables -- only for 2D image
+
+            # list of callables -- only for list of 2D or list of 3D images
             else:
                 _d = self.data
                 # only pass on kwargs if not None
@@ -699,7 +779,7 @@ def rgbplot(
     *,
     col_wrap=3,
     height=3,
-    aspect=1,
+    aspect="auto",
     cmap=None,
     alpha=None,
     origin=None,
@@ -715,6 +795,7 @@ def rgbplot(
     cbar_ticks=None,
     showticks=False,
     despine=None,
+    extent=None,
 ):
     """Split and plot the red, green and blue channels of an
     RGB image.
@@ -727,8 +808,8 @@ def rgbplot(
         Number of columns to display. Defaults to 3.
     height : int or float, optional
         Size of the individual images. Defaults to 3.
-    aspect : int or float, optional
-        Aspect ratio of individual images. Defaults to 1.
+    aspect : int, float or 'auto', optional
+        Aspect ratio of individual images, when set to 'auto', it calculates the aspect ratio of the images passed. Defaults to 'auto'.
     cmap : str or `matplotlib.colors.Colormap` or list, optional
         Image colormap or a list of colormaps. Defaults to None.
     alpha : float or array-like, optional
@@ -782,6 +863,9 @@ def rgbplot(
     despine : bool, optional
         Remove axes spines from image axes as well as colorbar axes.
         Defaults to None.
+    extent : floats (left, right, bottom, top), optional
+        The bounding box in data coordinates that the image will fill.
+        The image is stretched individually along x and y to fill the box.
 
     Returns
     -------
@@ -834,14 +918,14 @@ def rgbplot(
     """
 
     if not data.ndim == 3:
-        raise ValueError("imput image must be a RGB image")
+        raise ValueError("input image must be a RGB image")
 
     if data.shape[-1] != 3:
         raise ValueError("input image must be a RGB image")
 
     # if no cmap, assign reds, greens and blues cmap
     if cmap is None:
-        cmap = ["Reds", "Greens", "Blues"]
+        cmap = ["R", "G", "B"]
 
     # split RGB channels
     _d = [data[:, :, 0], data[:, :, 1], data[:, :, 2]]
@@ -866,6 +950,7 @@ def rgbplot(
         cbar_ticks=cbar_ticks,
         showticks=showticks,
         despine=despine,
+        extent=extent,
     )
 
     return g
@@ -904,8 +989,8 @@ class ParamGrid(object):
         is not None and `row` is None. Defaults to None.
     height : int or float, optional
         Size of the individual images. Defaults to 3.
-    aspect : int or float, optional
-        Aspect ratio of individual images. Defaults to 1.
+    aspect : int, float or 'auto', optional
+        Aspect ratio of individual images, when set to 'auto', it calculates the aspect ratio of the images passed. Defaults to 'auto'.
     cmap : str or `matplotlib.colors.Colormap`, optional
         Image colormap. Defaults to None.
     alpha : float or array-like, optional
@@ -953,6 +1038,8 @@ class ParamGrid(object):
     despine : bool, optional
         Remove axes spines from image axes as well as colorbar axes.
         Defaults to None.
+    extent : list, optional
+        Coordinates where to plot this image.
     **kwargs : Additional parameters as keyword arguments to be passed to the underlying filter specified.
 
     Returns
@@ -1040,7 +1127,7 @@ class ParamGrid(object):
         col=None,
         col_wrap=None,
         height=3,
-        aspect=1,
+        aspect="auto",
         cmap=None,
         alpha=None,
         origin=None,
@@ -1057,9 +1144,9 @@ class ParamGrid(object):
         cbar_ticks=None,
         showticks=False,
         despine=None,
+        extent=None,
         **kwargs,
     ):
-
         if data is None:
             raise ValueError("image data can not be None")
 
@@ -1100,6 +1187,9 @@ class ParamGrid(object):
             ncol = col_wrap
             nrow = int(np.ceil(len(kwargs[f"{col}"]) / col_wrap))
 
+        if aspect == "auto":
+            aspect = data.shape[1]/data.shape[0]
+
         # Calculate the base figure size
         figsize = (ncol * height * aspect, nrow * height)
 
@@ -1163,6 +1253,7 @@ class ParamGrid(object):
         self.cbar_ticks = cbar_ticks
         self.showticks = showticks
         self.despine = despine
+        self.extent = extent
 
         self._nrow = nrow
         self._ncol = ncol
@@ -1252,6 +1343,7 @@ class ParamGrid(object):
             cbar_ticks=self.cbar_ticks,
             showticks=self.showticks,
             despine=self.despine,
+            extent=self.extent,
             **func_kwargs,
         )
         return
